#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from doorway.x._uri import (
    UriMalformedException,
    UriIsIncorrectTypeError,
    UriValMode,
    UriTypeEnum,
    UriFieldValidator,
    UriValidator,
    UriValidatorUrl,
    UriValidatorFile,
    uri_parse,
    uri_validate,
    uri_extract,
    Uri,
)
from doorway.x._proxy import (
    proxies_set_default_scraper,
    proxies_register_scraper,
    proxies_scrape,
    ProxiesRunOutError,
    ProxyMalformedError,
    ProxyDownloadFailedError,
    ProxyRegisteredScraperHint,
    ProxyTypeHint,
    ProxyDictHint,
    ProxyScrapeFnHint,
    proxy_download,
    ProxyDownloader,
)

__all__ = [
    # ===== proxy ===== #
    # proxy scraper registration
    "proxies_set_default_scraper",
    "proxies_register_scraper",
    "proxies_scrape",
    # errors
    "ProxiesRunOutError",
    "ProxyMalformedError",
    "ProxyDownloadFailedError",
    # hints
    "ProxyRegisteredScraperHint",
    "ProxyTypeHint",
    "ProxyDictHint",
    "ProxyScrapeFnHint",
    # helpers
    "proxy_download",
    # main api
    "ProxyDownloader",
    # ===== uri ===== #
    # errors
    "UriMalformedException",
    "UriIsIncorrectTypeError",
    # enums
    "UriValMode",
    "UriTypeEnum",
    # validation
    "UriFieldValidator",
    "UriValidator",
    "UriValidatorUrl",
    "UriValidatorFile",
    # functional
    "uri_parse",
    "uri_validate",
    "uri_extract",
    # oop
    "Uri",
]



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


__all__ = (
    # errors
    "UriMalformedException",
    "UriIsIncorrectTypeError",
    # enums
    "UriValMode",
    "UriTypeEnum",
    # validation
    "UriFieldValidator",
    "UriValidator",
    "UriValidatorUrl",
    "UriValidatorFile",
    # functional
    "uri_parse",
    "uri_validate",
    "uri_extract",
    # oop
    "Uri",
)


import logging
import os
from contextlib import contextmanager
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

try:
    from rfc3986 import normalizers
    from rfc3986 import ParseResult
except ImportError:
    raise ImportError(
        "The `rfc3986` package is required for this module. "
        "You can install it via: `pip install rfc3986`."
    )


LOG = logging.getLogger(__name__)


# ========================================================================= #
# PATCH                                                                     #
# ========================================================================= #


_ORIG_REMOVE_DOT_SEGMENTS = normalizers.remove_dot_segments


@contextmanager
def _rfc3986_patch_context__remove_dot_segments(
    disabled: bool = False,
) -> Iterator[None]:
    # set new function, this no longer matches: http://tools.ietf.org/html/rfc3986#section-5.2.4
    # -- make sure that '..' and '.' at the start of a path are not removed!
    # -- '' might become '.' which should actually not be allowed!
    if not disabled:
        normalizers.remove_dot_segments = os.path.normpath
    # move into context
    try:
        yield
    # restore original function
    finally:
        normalizers.remove_dot_segments = _ORIG_REMOVE_DOT_SEGMENTS


# ========================================================================= #
# URI validation                                                            #
# ========================================================================= #


class UriMalformedException(Exception):
    def __init__(self, parsed: ParseResult, msg: str):
        super().__init__(f"[Malformed URI]: {msg} -- From URI: {parsed.unsplit()}")


class UriValMode(Enum):
    OPTIONAL = "OPTIONAL"
    REQUIRED = "REQUIRED"
    FORBIDDEN = "FORBIDDEN"


class UriFieldValidator:
    def __init__(
        self,
        mode: UriValMode = UriValMode.OPTIONAL,
        validator: Callable[[ParseResult, str, str, Any], ParseResult] = None,
        one_of: Optional[Sequence[Any]] = None,
    ):
        self._mode = mode
        self._validator = validator
        self._one_of = one_of

    def __call__(
        self, parsed: ParseResult, uri_kind: str, field_name: str, field_value: Any
    ) -> None:
        # validate based on the mode
        if self._mode == UriValMode.REQUIRED:
            if not field_value:
                raise UriMalformedException(
                    parsed,
                    f"field {repr(field_name)} is required, but got value: {repr(field_value)}",
                )
        elif self._mode == UriValMode.FORBIDDEN:
            if field_value:
                raise UriMalformedException(
                    parsed,
                    f"field {repr(field_name)} is forbidden, but got value: {repr(field_value)}",
                )
        elif self._mode != UriValMode.OPTIONAL:
            raise NotImplementedError("This should never happen!")
        # validate based on required values
        if self._one_of is not None:
            if field_value not in self._one_of:
                raise UriMalformedException(
                    parsed,
                    f"field {repr(field_name)} has value: {repr(field_value)}, but must be one of: {list(self._one_of)}",
                )
        # validate based on the validator function
        if self._validator is not None:
            self._validator(parsed, uri_kind, field_name, field_value)


class UriValidator:
    # override these in subclasses
    validate_scheme: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_host: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_port: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_path: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_query: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)

    def __call__(self, uri: Union[str, Path]) -> ParseResult:
        return self.validate(uri)

    def validate(self, uri: Union[str, Path]) -> ParseResult:
        parsed = uri_parse(uri)
        # validate everything
        self.validate_scheme(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="scheme",
            field_value=parsed.scheme,
        )
        self.validate_userinfo(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="userinfo",
            field_value=parsed.userinfo,
        )
        self.validate_host(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="host",
            field_value=parsed.host,
        )
        self.validate_port(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="port",
            field_value=parsed.port,
        )
        self.validate_path(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="path",
            field_value=parsed.path,
        )
        self.validate_query(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="query",
            field_value=parsed.query,
        )
        self.validate_fragment(
            parsed=parsed,
            uri_kind=self.uri_kind,
            field_name="fragment",
            field_value=parsed.fragment,
        )
        # final result
        return parsed

    # override these

    @property
    def uri_kind(self) -> str:
        raise NotImplementedError

    @property
    def uri_type(self) -> "UriTypeEnum":
        raise NotImplementedError

    @property
    def allowed_schemes(self) -> Tuple[Optional[str], ...]:
        raise NotImplementedError

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        raise NotImplementedError


# ========================================================================= #
# URI Types                                                                 #
# ========================================================================= #


# including `str` as a parent class means we can write: `assert UriType.URL == 'URL'`
# without `str` as a parent class, this does not evaluate to `True`.
# -- HOWEVER: we do not want this feature! It will just cause confusion
#             and can't be type checked!
class UriTypeEnum(Enum):
    """
    Types of URIs supported by the `parse_uri_and_type` function.
    """

    FILE = "FILE"
    URL = "URL"
    S3 = "S3"
    SSH = "SSH"


class UriValidatorUrl(UriValidator):
    uri_kind = "url"
    uri_type = UriTypeEnum.URL
    allowed_schemes = ("http", "https")

    # override these in subclasses
    validate_scheme: UriFieldValidator = UriFieldValidator(
        mode=UriValMode.REQUIRED, one_of=allowed_schemes
    )
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)
    validate_host: UriFieldValidator = UriFieldValidator(mode=UriValMode.REQUIRED)
    validate_port: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_path: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_query: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=UriValMode.OPTIONAL)

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        return validated.geturl()


class UriValidatorFile(UriValidator):
    uri_kind = "file"
    uri_type = UriTypeEnum.FILE
    allowed_schemes = ("file", None)

    # override these in subclasses
    validate_scheme: UriFieldValidator = UriFieldValidator(
        mode=UriValMode.OPTIONAL, one_of=allowed_schemes
    )
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)
    validate_host: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)
    validate_port: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)
    validate_path: UriFieldValidator = UriFieldValidator(mode=UriValMode.REQUIRED)
    validate_query: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=UriValMode.FORBIDDEN)

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        return validated.path


# ========================================================================= #
# URI vars                                                                  #
# ========================================================================= #


_SCHEME_VALIDATORS: Dict[Optional[str], UriValidator] = {
    scheme: validate_cls()
    for validate_cls in [UriValidatorUrl, UriValidatorFile]
    for scheme in validate_cls.allowed_schemes
}


# ========================================================================= #
# URI validation                                                            #
# ========================================================================= #


def uri_parse(
    uri: Union[str, Path, ParseResult], rfc3986_norm: bool = False
) -> ParseResult:
    parsed = uri
    # convert to parse result
    # -- assume already normalized if already a ParseResult
    if not isinstance(parsed, ParseResult):
        with _rfc3986_patch_context__remove_dot_segments(disabled=rfc3986_norm):
            parsed: ParseResult = ParseResult.from_string(
                str(parsed), lazy_normalize=False
            )
    # done!
    return parsed


def uri_validate(
    uri: Union[str, Path], return_validator: bool = False
) -> Union[ParseResult, Tuple[ParseResult, UriValidator]]:
    parsed = uri_parse(uri)
    # get the validator
    validator = _SCHEME_VALIDATORS.get(parsed.scheme, None)
    if validator is None:
        raise KeyError(
            f"invalid uri scheme: {repr(parsed.scheme)}, must be one of: {list(_SCHEME_VALIDATORS.keys())}, for uri: {repr(parsed.geturl())}"
        )
    # validate the uri
    validated = validator.validate(parsed)
    # get results
    if return_validator:
        return validated, validator
    return validated


def uri_extract(
    uri: Union[str, Path],
    return_validated: bool = False,
    return_validator: bool = False,
) -> Union[
    str,
    Tuple[str, ParseResult],
    Tuple[str, UriValidator],
    Tuple[str, ParseResult, UriValidator],
]:
    validated, validator = uri_validate(uri, return_validator=True)
    # validate the uri
    uri_norm = validator.extract(validated)
    # return the single result
    if not (return_validator or return_validated):
        return uri_norm
    # return all the results
    results = [uri_norm]
    if return_validated:
        results.append(validated)
    if return_validator:
        results.append(validator)
    return tuple(results)


# ========================================================================= #
# URI Class - Errors                                                        #
# ========================================================================= #


T = TypeVar("T")


class UriIsIncorrectTypeError(Exception):
    """
    This error is thrown if the uri is an incorrect type
    """


def _only_if(prop: property) -> Callable[[T], T]:
    def decorator(func: T) -> T:
        @wraps(func)
        def wrapper(self: "Uri", *args, **kwargs):
            if getattr(self, prop.fget.__name__):
                raise UriIsIncorrectTypeError(
                    f"Check if: `{prop.fget.__name__}` is `True` before calling `{func.__name__}`, got uri: {repr(self.uri)}"
                )
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


# ========================================================================= #
# URI Class                                                                 #
# ========================================================================= #


class Uri(object):
    def __init__(self, uri: Union[str, Path, ParseResult, "Uri"]):
        # unwrap uri object
        if isinstance(uri, Uri):
            uri = uri._input_uri
            assert not isinstance(uri, Uri)
        # save the input
        self._input_uri: Union[str, Path, ParseResult] = uri
        # get validated uri
        validated, validator = uri_validate(uri, return_validator=True)
        self._validated: ParseResult = validated
        self._validator: UriValidator = validator

    # ~=~=~ URI ~=~=~ #

    @property
    def uri_extract(self) -> str:
        return self._validator.extract(self._validated)

    @property
    def uri_type(self) -> UriTypeEnum:
        return self._validator.uri_type

    @property
    def uri_parsed(self) -> ParseResult:
        return self._validated

    @property
    def uri_basename(self) -> str:
        return os.path.basename(self._validated.path)

    @property
    def uri(self) -> str:
        return self._validated.geturl()

    def __repr__(self):
        return f"{self.__class__.__name__}(uri={repr(self._input_uri)})"

    def __str__(self):
        return self.uri

    # ~=~=~ FILE ~=~=~ #

    @property
    def is_file(self) -> bool:
        return self.uri_type == UriTypeEnum.FILE

    @property
    @_only_if(is_file)
    def file(self) -> str:
        return self.uri_extract

    @property
    @_only_if(is_file)
    def file_abs(self) -> str:
        return os.path.abspath(self.uri_extract)

    @property
    @_only_if(is_file)
    def file_is_abs(self) -> bool:
        return os.path.isabs(self.uri_extract)

    # ~=~=~ URL ~=~=~ #

    @property
    def is_url(self) -> bool:
        return self.uri_type == UriTypeEnum.URL

    @property
    @_only_if(is_url)
    def url(self) -> str:
        return self.uri_extract

    @property
    @_only_if(is_url)
    def url_is_http(self) -> bool:
        return self.uri_parsed.scheme == "http"

    @property
    @_only_if(is_url)
    def url_is_https(self) -> bool:
        return self.uri_parsed.scheme == "https"

    # ~=~=~ S3 ~=~=~ #

    @property
    def is_s3(self) -> bool:
        return self.uri_type == UriTypeEnum.S3

    # TODO ...

    # ~=~=~ SSH ~=~=~ #

    @property
    def is_ssh(self) -> bool:
        return self.uri_type == UriTypeEnum.SSH

    # TODO ...

    # ~=~=~ COMMON ~=~=~ #

    # def download(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError
    #
    # def copy(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError
    #
    # def retrieve(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
import pytest

from doorway._fmt import fmt_bytes_to_human
from doorway._ctx import ctx_temp_environ


# ========================================================================= #
# TEST UTILS                                                                #
# ========================================================================= #


def test_fmt_bytes_to_human():
    with ctx_temp_environ(DOORWAY_ENABLE_COLORS="false"):
        assert fmt_bytes_to_human(1000**1, base=1024, align=True) == "1000.000 B  "
        assert fmt_bytes_to_human(1000**2, base=1024, align=True) == " 976.562 KiB"
        assert fmt_bytes_to_human(1000**3, base=1024, align=True) == " 953.674 MiB"
        assert fmt_bytes_to_human(1000**4, base=1024, align=True) == " 931.323 GiB"
        assert fmt_bytes_to_human(1000**5, base=1024, align=True) == " 909.495 TiB"

        assert fmt_bytes_to_human(1024**1, base=1024, align=True) == "   1.000 KiB"
        assert fmt_bytes_to_human(1024**2, base=1024, align=True) == "   1.000 MiB"
        assert fmt_bytes_to_human(1024**3, base=1024, align=True) == "   1.000 GiB"
        assert fmt_bytes_to_human(1024**4, base=1024, align=True) == "   1.000 TiB"
        assert fmt_bytes_to_human(1024**5, base=1024, align=True) == "   1.000 PiB"

        assert fmt_bytes_to_human(3070**1, base=1024, align=True) == "   2.998 KiB"
        assert fmt_bytes_to_human(3070**2, base=1024, align=True) == "   8.988 MiB"
        assert fmt_bytes_to_human(3070**3, base=1024, align=True) == "  26.947 GiB"
        assert fmt_bytes_to_human(3070**4, base=1024, align=True) == "  80.789 TiB"
        assert fmt_bytes_to_human(3070**5, base=1024, align=True) == " 242.210 PiB"

        assert fmt_bytes_to_human(1000**1, base=1024) == "1000.000 B"
        assert fmt_bytes_to_human(1000**2, base=1024) == "976.562 KiB"
        assert fmt_bytes_to_human(1000**3, base=1024) == "953.674 MiB"
        assert fmt_bytes_to_human(1000**4, base=1024) == "931.323 GiB"
        assert fmt_bytes_to_human(1000**5, base=1024) == "909.495 TiB"

        assert fmt_bytes_to_human(1024**1, base=1024) == "1.000 KiB"
        assert fmt_bytes_to_human(1024**2, base=1024) == "1.000 MiB"
        assert fmt_bytes_to_human(1024**3, base=1024) == "1.000 GiB"
        assert fmt_bytes_to_human(1024**4, base=1024) == "1.000 TiB"
        assert fmt_bytes_to_human(1024**5, base=1024) == "1.000 PiB"

        assert fmt_bytes_to_human(3070**1, base=1024) == "2.998 KiB"
        assert fmt_bytes_to_human(3070**2, base=1024) == "8.988 MiB"
        assert fmt_bytes_to_human(3070**3, base=1024) == "26.947 GiB"
        assert fmt_bytes_to_human(3070**4, base=1024) == "80.789 TiB"
        assert fmt_bytes_to_human(3070**5, base=1024) == "242.210 PiB"

        assert fmt_bytes_to_human(999, base=1024) == "999.000 B"
        assert fmt_bytes_to_human(1000, base=1024) == "1000.000 B"
        assert fmt_bytes_to_human(1001, base=1024) == "1001.000 B"
        assert fmt_bytes_to_human(1999, base=1024) == "1.952 KiB"
        assert fmt_bytes_to_human(2000, base=1024) == "1.953 KiB"
        assert fmt_bytes_to_human(2001, base=1024) == "1.954 KiB"

        assert fmt_bytes_to_human(1023, base=1024) == "1023.000 B"
        assert fmt_bytes_to_human(1024, base=1024) == "1.000 KiB"
        assert fmt_bytes_to_human(1025, base=1024) == "1.001 KiB"
        assert fmt_bytes_to_human(2047, base=1024) == "1.999 KiB"
        assert fmt_bytes_to_human(2048, base=1024) == "2.000 KiB"
        assert fmt_bytes_to_human(2049, base=1024) == "2.001 KiB"

        assert fmt_bytes_to_human(999, base=1000) == "999.000 B"
        assert fmt_bytes_to_human(1000, base=1000) == "1.000 kB"
        assert fmt_bytes_to_human(1001, base=1000) == "1.001 kB"
        assert fmt_bytes_to_human(1999, base=1000) == "1.999 kB"
        assert fmt_bytes_to_human(2000, base=1000) == "2.000 kB"
        assert fmt_bytes_to_human(2001, base=1000) == "2.001 kB"

        assert fmt_bytes_to_human(1023, base=1000) == "1.023 kB"
        assert fmt_bytes_to_human(1024, base=1000) == "1.024 kB"
        assert fmt_bytes_to_human(1025, base=1000) == "1.025 kB"
        assert fmt_bytes_to_human(2047, base=1000) == "2.047 kB"
        assert fmt_bytes_to_human(2048, base=1000) == "2.048 kB"
        assert fmt_bytes_to_human(2049, base=1000) == "2.049 kB"

        assert fmt_bytes_to_human(2_001_000, base=1000) == "2.001 MB"

        assert fmt_bytes_to_human(1025**0, base=1024) == "1.000 B"  # should be "1 B"?
        assert fmt_bytes_to_human(1025**0, base=1000) == "1.000 B"  # should be "1 B"?
        assert fmt_bytes_to_human(1025**1, base=1024) == "1.001 KiB"
        assert fmt_bytes_to_human(1025**1, base=1000) == "1.025 kB"
        assert fmt_bytes_to_human(1025**2, base=1024) == "1.002 MiB"
        assert fmt_bytes_to_human(1025**2, base=1000) == "1.051 MB"
        assert fmt_bytes_to_human(1025**3, base=1024) == "1.003 GiB"
        assert fmt_bytes_to_human(1025**3, base=1000) == "1.077 GB"
        assert fmt_bytes_to_human(1025**4, base=1024) == "1.004 TiB"
        assert fmt_bytes_to_human(1025**4, base=1000) == "1.104 TB"
        assert fmt_bytes_to_human(1025**5, base=1024) == "1.005 PiB"
        assert fmt_bytes_to_human(1025**5, base=1000) == "1.131 PB"
        assert fmt_bytes_to_human(1025**6, base=1024) == "1.006 EiB"
        assert fmt_bytes_to_human(1025**6, base=1000) == "1.160 EB"
        assert fmt_bytes_to_human(1025**7, base=1024) == "1.007 ZiB"
        assert fmt_bytes_to_human(1025**7, base=1000) == "1.189 ZB"
        assert fmt_bytes_to_human(1025**8, base=1024) == "1.008 YiB"
        assert fmt_bytes_to_human(1025**8, base=1000) == "1.218 YB"
        assert fmt_bytes_to_human(1025**9, base=1024) == "1.009 RiB"
        assert fmt_bytes_to_human(1025**9, base=1000) == "1.249 RB"
        assert fmt_bytes_to_human(1025**10, base=1024) == "1.010 QiB"
        assert fmt_bytes_to_human(1025**10, base=1000) == "1.280 QB"

        # check out of valid range of units
        assert fmt_bytes_to_human(1025**11, base=1024) == "1035.054 QiB"
        assert fmt_bytes_to_human(1025**11, base=1000) == "1312.087 QB"
        assert fmt_bytes_to_human(1025**12, base=1024) == "1060930.215 QiB"
        assert fmt_bytes_to_human(1025**12, base=1000) == "1344888.824 QB"

        with pytest.raises(
            ValueError, match="invalid size in bytes, cannot be negative: -1337"
        ):
            fmt_bytes_to_human(-1337, base=1000)

        # check rounding with +1 or -1 values

        assert fmt_bytes_to_human(1024**2 - 1, base=1024) == "1023.999 KiB"
        assert fmt_bytes_to_human(1024**2 + 0, base=1024) == "1.000 MiB"
        assert fmt_bytes_to_human(1024**2 + 1, base=1024) == "1.000 MiB"

        assert (
            fmt_bytes_to_human(1024**3 - 1, base=1024, round_unit=False)
            == "1024.000 MiB"
        )
        assert fmt_bytes_to_human(1024**3 - 1, base=1024) == "1.000 GiB"
        assert fmt_bytes_to_human(1024**3 + 0, base=1024) == "1.000 GiB"
        assert fmt_bytes_to_human(1024**3 + 1, base=1024) == "1.000 GiB"

        assert fmt_bytes_to_human(1000**3 - 1, base=1024) == "953.674 MiB"
        assert fmt_bytes_to_human(1000**3 + 0, base=1024) == "953.674 MiB"
        assert fmt_bytes_to_human(1000**3 + 1, base=1024) == "953.674 MiB"

        assert fmt_bytes_to_human(1024**3 - 1, base=1000) == "1.074 GB"
        assert fmt_bytes_to_human(1024**3 + 0, base=1000) == "1.074 GB"
        assert fmt_bytes_to_human(1024**3 + 1, base=1000) == "1.074 GB"

        assert (
            fmt_bytes_to_human(1000**3 - 1, base=1000, round_unit=False)
            == "1000.000 MB"
        )
        assert fmt_bytes_to_human(1000**3 - 1, base=1000) == "1.000 GB"
        assert fmt_bytes_to_human(1000**3 + 0, base=1000) == "1.000 GB"
        assert fmt_bytes_to_human(1000**3 + 1, base=1000) == "1.000 GB"

        # check rounding styles

        assert (
            fmt_bytes_to_human(1024**3 - 1, base=1024, round_unit=True) == "1.000 GiB"
        )
        assert (
            fmt_bytes_to_human(1024**3 - 1, base=1024, round_unit=False)
            == "1024.000 MiB"
        )
        assert (
            fmt_bytes_to_human(1024**3 - 524, base=1024, round_unit=True) == "1.000 GiB"
        )
        assert (
            fmt_bytes_to_human(1024**3 - 524, base=1024, round_unit=False)
            == "1024.000 MiB"
        )
        assert (
            fmt_bytes_to_human(1024**3 - 525, base=1024, round_unit=True)
            == "1023.999 MiB"
        )
        assert (
            fmt_bytes_to_human(1024**3 - 525, base=1024, round_unit=False)
            == "1023.999 MiB"
        )

        assert fmt_bytes_to_human(1000**3 - 1, base=1000, round_unit=True) == "1.000 GB"
        assert (
            fmt_bytes_to_human(1000**3 - 1, base=1000, round_unit=False)
            == "1000.000 MB"
        )
        assert (
            fmt_bytes_to_human(1000**3 - 500, base=1000, round_unit=True) == "1.000 GB"
        )
        assert (
            fmt_bytes_to_human(1000**3 - 500, base=1000, round_unit=False)
            == "1000.000 MB"
        )
        assert (
            fmt_bytes_to_human(1000**3 - 501, base=1000, round_unit=True)
            == "999.999 MB"
        )
        assert (
            fmt_bytes_to_human(1000**3 - 501, base=1000, round_unit=False)
            == "999.999 MB"
        )

        # check default values
        assert fmt_bytes_to_human(1000**4, base=1024) == fmt_bytes_to_human(1000**4)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

__all__ = [
    # proxy scraper registration
    "proxies_set_default_scraper",
    "proxies_register_scraper",
    "proxies_scrape",
    # errors
    "ProxiesRunOutError",
    "ProxyMalformedError",
    "ProxyDownloadFailedError",
    # hints
    "ProxyRegisteredScraperHint",
    "ProxyTypeHint",
    "ProxyDictHint",
    "ProxyScrapeFnHint",
    # helpers
    "proxy_download",
    # main api
    "ProxyDownloader",
]

import io
import os
from collections import defaultdict
from logging import getLogger
from multiprocessing.pool import ThreadPool
from pathlib import Path
from random import Random
from typing import Callable, Iterable, Literal, TYPE_CHECKING, TypedDict, TypeVar, Union
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
import urllib.request

from doorway import EnvVar

_LOGGER = getLogger(__name__)


if TYPE_CHECKING:
    import requests


# ============================================================================ #
# Proxy Scraping                                                               #
# ============================================================================ #


class HTTPProxyHint(TypedDict):
    HTTP: str


class HTTPSProxyHint(TypedDict):
    HTTPS: str


ProxyRegisteredScraperHint = str
ProxyTypeHint = Literal["http", "https", "all"]
ProxyDictHint = Union[HTTPProxyHint, HTTPSProxyHint]
ProxyScrapeFnHint = Callable[[ProxyTypeHint], List[ProxyDictHint]]

_GenericScrapeFn = TypeVar("_GenericScrapeFn", bound=ProxyScrapeFnHint)


# ============================================================================ #
# Proxy Scraping                                                               #
# ============================================================================ #


_PROXY_SOURCES = {}
_DEFAULT_SOURCE = None


def proxies_set_default_scraper(name: ProxyRegisteredScraperHint) -> None:
    if name not in _PROXY_SOURCES:
        raise KeyError(
            f"Cannot set as default! Scrape function with name: {repr(name)} does not exist."
        )
    global _DEFAULT_SOURCE
    if _DEFAULT_SOURCE != name:
        _LOGGER.info(
            f"overridden default proxy scrape_fn: {repr(_DEFAULT_SOURCE)} -> {repr(name)}"
        )
        _DEFAULT_SOURCE = name


def proxies_register_scraper(
    name: ProxyRegisteredScraperHint,
    *,
    is_default: bool = False,
    scrape_fn: Optional[_GenericScrapeFn] = None,
) -> Union[_GenericScrapeFn, Callable[[_GenericScrapeFn], _GenericScrapeFn]]:
    if name in _PROXY_SOURCES:
        raise KeyError("scrape function with name: {repr(name)} already exists")

    # decorator
    def wrapper(scrape_fn: _GenericScrapeFn) -> _GenericScrapeFn:
        # just in case the decorator call was delayed
        assert name not in _PROXY_SOURCES
        _PROXY_SOURCES.setdefault(name, scrape_fn)
        _LOGGER.debug(f"registered proxy scrape_fn: {repr(name)}")
        # set the default
        if is_default:
            proxies_set_default_scraper(name=name)
        return scrape_fn

    # decorator or function
    if scrape_fn is None:
        return wrapper
    else:
        return wrapper(scrape_fn)


def proxies_scrape(
    source: Optional[str] = ProxyRegisteredScraperHint,
    proxy_type: ProxyTypeHint = "all",
    cache_dir: str = "data/proxies/cachier",
    cached: bool = True,
) -> List[ProxyDictHint]:
    if source is None:
        if _DEFAULT_SOURCE is None:
            raise RuntimeError("no default proxy scrape function has been set.")
        source = _DEFAULT_SOURCE
        _LOGGER.info(f"using default proxy scrape function: {repr(source)}")
    # get source
    try:
        proxy_scrape_fn = _PROXY_SOURCES[source]
    except KeyError:
        raise KeyError(
            f"proxy scrape function with name: {repr(source)} does not exist. Valid scrape sources are: {sorted(_PROXY_SOURCES.keys())}"
        )
    # wrap the function
    if cached:
        try:
            from cachier import cachier
        except ImportError as e:
            raise ImportError(
                "To use the 'cached' feature, you must install cachier."
                "You can install it via: `pip install cachier`"
            ) from e

        from datetime import timedelta

        proxy_scrape_fn = cachier(
            stale_after=timedelta(days=1), backend="pickle", cache_dir=cache_dir
        )(proxy_scrape_fn)
    # obtain the proxies
    _LOGGER.info(f"scraping proxies from source: {repr(source)}")
    proxy_list = proxy_scrape_fn(proxy_type=proxy_type)
    _LOGGER.info(f"scrapped: {len(proxy_list)} proxies from source: {repr(source)}")
    # done!
    return proxy_list


# ============================================================================ #
# Proxy Scrapers                                                               #
# ============================================================================ #


def _requests_get(
    url: str,
    fake_user_agent: bool = True,
    params: Optional[dict] = None,
) -> "requests.Response":
    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "To use the 'requests' library, you must install it."
            "You can install it via: `pip install requests`"
        ) from e

    # fake a request from a browser
    return requests.get(
        url,
        headers={}
        if not fake_user_agent
        else {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.212 Safari/537.36"
        },
        params=params,
    )


@proxies_register_scraper("proxylist.geonode.com", is_default=True)
def _scrape_proxylist_geonode_com(proxy_type: ProxyTypeHint) -> List[ProxyDictHint]:
    def _get_page(page):
        r = _requests_get(
            f"https://proxylist.geonode.com/api/proxy-list?limit=500&page={page}&sort_by=lastChecked&sort_type=desc",
            fake_user_agent=True,
        )
        r.raise_for_status()
        return r.json()

    proxies: "List[ProxyDictHint]" = []
    page_num = 1
    while True:
        print(f"page_num={page_num}")
        data = _get_page(page_num)
        for row in data["data"]:
            proto = row["protocols"][0].upper()
            url = f"{proto}://{row['ip']}:{row['port']}"
            proxies.append({proto: url})
        if page_num >= data["total"]:
            break
        page_num += 1
        if page_num > 3:
            break

    # filter by type
    if proxy_type == "http":
        proxies = [p for p in proxies if "HTTP" in p]
    elif proxy_type == "https":
        proxies = [p for p in proxies if "HTTPS" in p]
    else:
        pass

    return proxies


@proxies_register_scraper("morph.io")
def _scrape_proxies_morph(proxy_type: ProxyTypeHint) -> List[ProxyDictHint]:
    morph_api_key = EnvVar.env_str("MORPH_API_KEY").get()
    morph_api_url = "https://api.morph.io/CookieMichal/us-proxy/data.json"

    query = (
        "select * from 'data' where (anonymity='elite proxy' or anonymity='anonymous')"
    )

    if "https" == proxy_type:
        query += " and https='yes'"
    elif "http" == proxy_type:
        query += " and https='no'"
    elif "all" == proxy_type:
        pass
    else:
        raise KeyError(f"invalid proxy_type: {proxy_type}")

    r = _requests_get(morph_api_url, params={"key": morph_api_key, "query": query})

    proxies: "List[ProxyDictHint]" = []
    for row in r.json():
        proto = "HTTPS" if row["https"] == "yes" else "HTTP"
        url = "{}://{}:{}".format(proto, row["ip"], row["port"])
        proxies.append({proto: url})

    return proxies


@proxies_register_scraper("free-proxy-list.net")
def _scrape_proxies_freeproxieslist(proxy_type: ProxyTypeHint) -> List[Dict[str, str]]:
    def can_add(https):
        if proxy_type == "all":
            return True
        elif proxy_type == "https":
            return https == "yes"
        elif proxy_type == "http":
            return https == "no"
        else:
            raise KeyError(f"invalid proxy_type: {proxy_type}")

    try:
        from bs4 import BeautifulSoup
    except ImportError as e:
        raise ImportError(
            "To use the 'free-proxy-list.net' scraper, you must install bs4."
            "You can install it via: `pip install beautifulsoup4`"
        ) from e

    page = _requests_get("https://free-proxy-list.net/", fake_user_agent=True)
    soup = BeautifulSoup(page.content, "html.parser")
    rows = soup.find_all("tr", recursive=True)

    proxies: "List[ProxyDictHint]" = []
    for row in rows:
        try:
            ip, port, country, country_long, anonymity, google, https, last_checked = (
                elem.text for elem in row.find_all("td", recursive=True)
            )
            # check this entry is an ip entry
            if len(ip.split(".")) != 4:
                raise ValueError("not an ip entry")
            # filter entries
            if not can_add(https):
                continue
            # make entry
            proto = "HTTPS" if (https == "yes") else "HTTP"
            url = "{}://{}:{}".format(proto, ip, int(port))
            proxies.append({proto: url})
        except Exception:
            pass

    return proxies


# ============================================================================ #
# Proxy Errors                                                                 #
# ============================================================================ #


class ProxyMalformedError(Exception):
    """
    raised if a proxy does not follow the correct format.
    eg. `proxy = {'<protocol>': '<protocol>://<url>'}`
    """


class ProxiesRunOutError(Exception):
    """
    raise if the ProxyDownloader has run out of proxies!
    """


class ProxyDownloadFailedError(Exception):
    """
    raise if the ProxyDownloader download has failed
    """


# ============================================================================ #
# Proxy Download Helper                                                        #
# ============================================================================ #


def _make_proxy_opener(proxy: ProxyDictHint):
    if len(proxy) != 1:
        raise ProxyMalformedError(
            f"proxy dictionaries should only have one entry, the key is the protocol, and the value is the url... invalid: {proxy}"
        )
    # build connection
    return urllib.request.build_opener(
        urllib.request.ProxyHandler(proxy), urllib.request.ProxyBasicAuthHandler()
    )


def proxy_download(
    url: str,
    file: Union[str, Path],
    proxy: ProxyDictHint,
    timeout: Optional[float] = 8,
):
    # TODO: should use AtomicOpen
    data = _make_proxy_opener(proxy=proxy).open(url, timeout=timeout).read()
    # download to temp file in case there is an error
    file = str(file)
    temp_file = file + ".dl"
    with io.FileIO(temp_file, "w") as f:
        f.write(data)
    # make this atomic
    os.rename(temp_file, file)


def _skip_or_prepare_file(
    file: Union[str, Path],
    exists_mode: Literal["error", "skip", "overwrite"],
    make_dirs: bool,
):
    """
    returns True if the file should be skipped, False otherwise.
    - also prepare the directories or deletion of the file!
    """
    file = str(file)

    if os.path.exists(file):
        # the file exists
        # make sure it is actually a file, not a directory or link
        if not os.path.isfile(file):
            raise IOError(f"the specified file is not a file: {file}")
        # handle the different modes
        if exists_mode == "error":
            raise FileExistsError(f"the file already exists: {file}")
        elif exists_mode == "skip":
            return True
        elif exists_mode == "overwrite":
            os.unlink(file)
            _LOGGER.warning("overwriting file: {url}")
        else:
            raise KeyError(f"invalid exists_mode={repr(exists_mode)}")
    else:
        # the file does not exist
        # check the parent path
        parent_dir = os.path.dirname(file)
        if not os.path.exists(parent_dir):
            # the parent path does not exist
            if make_dirs:
                os.makedirs(parent_dir, exist_ok=True)
                _LOGGER.debug(f"[MADE] directory: {parent_dir}")
            else:
                raise FileNotFoundError(
                    f"Parent directory does not exist: {parent_dir} Otherwise set make_dirs=True"
                )
        else:
            # the parent path exists
            if not os.path.isdir(parent_dir):
                raise NotADirectoryError(
                    f"Parent directory is not a directory: {parent_dir}"
                )
    return False


# ============================================================================ #
# Proxy Downloader                                                             #
# ============================================================================ #


class ProxyDownloader:
    def __init__(
        self,
        proxies: Optional[
            Union[Sequence[ProxyDictHint], ProxyRegisteredScraperHint]
        ] = None,
        req_min_remove_count: int = 5,
        req_max_fail_ratio: float = 0.5,
    ):
        # default proxy scraping
        if proxies is None:
            proxies = proxies_scrape()
        elif isinstance(proxies, str):
            proxies = proxies_scrape(source=proxies)
        # convert
        self._proxies = list(proxies)  # TODO: add support for raw proxy strings?
        # proxy statistics
        self._req_counts = defaultdict(int)
        self._req_fails = defaultdict(int)
        self._req_max_fail_ratio = req_max_fail_ratio
        self._req_min_remove_count = req_min_remove_count
        # random instance
        self._rand = Random()  # TODO: add round robbin mode?

    def random_proxy(self) -> ProxyDictHint:
        if len(self._proxies) <= 0:
            raise ProxiesRunOutError(
                "The proxy downloader has run out of valid proxies."
            )
        # return a random proxy!
        index = self._rand.randint(0, len(self._proxies) - 1)
        return self._proxies[index]

    def _update_proxy(self, proxy: ProxyDictHint, success: bool) -> None:
        (purl,) = proxy.values()
        # update uses and failures
        self._req_counts[purl] += 1
        self._req_fails[purl] += int(bool(not success))
        # make remove if there was an error
        counts, fails = self._req_counts[purl], self._req_fails[purl]
        if (counts > self._req_min_remove_count) and (
            fails / counts > self._req_max_fail_ratio
        ):
            try:
                self._proxies.remove(proxy)
                del self._req_counts[purl]
                del self._req_fails[purl]
            except (ValueError, KeyError):
                pass  # removed in another thread

    def download_threaded(
        self,
        url_file_tuples: Iterable[Tuple[str, Union[str, Path]]],
        exists_mode: Literal["error", "skip", "overwrite"] = "error",
        verbose: bool = False,
        make_dirs: bool = False,
        ignore_failures=False,
        threads=64,
        attempts: int = 128,
        timeout: int = 8,
    ):
        try:
            from tqdm import tqdm
        except ImportError:
            raise ImportError(
                "To use the download_threaded function, you must install tqdm."
                "You can install it via: `pip install tqdm`"
            )

        # check inputs
        try:
            total = len(url_file_tuples)
            if total <= 0:
                return []
        except TypeError:
            total = None

        def download(url_file):
            url, file = url_file
            try:
                self.download(
                    url=url,
                    file=file,
                    exists_mode=exists_mode,
                    verbose=verbose,
                    make_dirs=make_dirs,
                    attempts=attempts,
                    timeout=timeout,
                )
            except ProxyDownloadFailedError:
                if ignore_failures:
                    return url, file
                else:
                    raise
            return None

        def get_desc():
            if ignore_failures:
                return (
                    f"Downloading [p={len(self._proxies)},t={threads},f={len(failed)}]"
                )
            else:
                return f"Downloading [p={len(self._proxies)},t={threads}]"

        # download all files, keeping track of failed items!
        failed = []
        with ThreadPool(processes=threads) as pool:
            with tqdm(desc=get_desc(), total=total) as pbar:
                for pair in pool.imap_unordered(download, url_file_tuples):
                    if pair:
                        failed.append(pair)
                    pbar.desc = get_desc()
                    pbar.update()

        # return all tuples for failed attempts
        return failed

    def download(
        self,
        url: str,
        file: Union[str, Path],
        exists_mode: Literal["error", "skip", "overwrite"] = "error",
        verbose: bool = False,
        make_dirs: bool = False,
        attempts: int = 128,
        timeout: int = 8,
    ):
        """
        Download a file using random proxies.
        """
        if _skip_or_prepare_file(
            file=file, exists_mode=exists_mode, make_dirs=make_dirs
        ):
            if verbose:
                _LOGGER.debug(f"[SKIPPED]: {file} | {url}")
            return
        # attempt download
        for i in range(attempts):
            proxy = self.random_proxy()
            try:
                proxy_download(url, file, proxy=proxy, timeout=timeout)
                if verbose:
                    _LOGGER.info(f"[DOWNLOADED]: {file} | {url}")
                self._update_proxy(proxy, success=True)
                return
            except Exception as e:
                if verbose:
                    _LOGGER.debug(f"[FAILED ATTEMPT {i + 1}]: {file} | {url} -- {e}")
                self._update_proxy(proxy, success=False)
        # download failed
        raise ProxyDownloadFailedError(f"[FAILED] tries={attempts}: {file} | {url}")


# ============================================================================ #
# Entrypoint                                                                   #
# ============================================================================ #


if __name__ == "__main__":

    def _command_line_app():
        import argparse
        import logging

        # parse arguments
        parser = argparse.ArgumentParser()
        parser.add_argument("-d", "--cache-dir", type=str, default="data/cache/proxies")
        parser.add_argument("-t", "--proxy-type", type=str, default="all")
        parser.add_argument("-s", "--proxy-source", type=str, default=None)
        parser.add_argument("-f", "--force-download", action="store_true")
        args = parser.parse_args()

        # download the proxies
        logging.basicConfig(level=logging.DEBUG)
        ProxyDownloader(
            proxies=proxies_scrape(
                source=args.proxy_source,
                proxy_type=args.proxy_type,
                cache_dir=args.cache_dir,
                cached=not args.force_download,
            )
        )

    _command_line_app()


# ============================================================================ #
# END                                                                          #
# ============================================================================ #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import pytest

# ========================================================================= #
# TEST UTILS                                                                #
# ========================================================================= #


@pytest.mark.extra_test
def test_uri_type_enum():
    from doorway.x._uri import UriTypeEnum

    assert len(UriTypeEnum) == 4
    assert list(UriTypeEnum) == [
        UriTypeEnum.FILE,
        UriTypeEnum.URL,
        UriTypeEnum.S3,
        UriTypeEnum.SSH,
    ]
    # test types
    assert isinstance(UriTypeEnum.FILE, UriTypeEnum)
    assert isinstance(UriTypeEnum.FILE.value, str)
    assert isinstance(UriTypeEnum.FILE.name, str)
    # make sure the inverse is not true
    assert not isinstance(UriTypeEnum.FILE, str)
    assert not isinstance(UriTypeEnum.FILE.value, UriTypeEnum)
    assert not isinstance(UriTypeEnum.FILE.name, UriTypeEnum)
    # test equivalence
    assert UriTypeEnum.FILE == UriTypeEnum.FILE
    assert UriTypeEnum.FILE != UriTypeEnum.URL
    assert UriTypeEnum.URL == UriTypeEnum.URL
    # check enums are not equal to their values
    assert UriTypeEnum.FILE != UriTypeEnum.FILE.value
    assert UriTypeEnum.FILE != UriTypeEnum.FILE.name
    assert UriTypeEnum.URL != UriTypeEnum.URL.value
    assert UriTypeEnum.URL != UriTypeEnum.URL.name
    # check names and values are equivalent
    assert UriTypeEnum.FILE.name == UriTypeEnum.FILE.value
    assert UriTypeEnum.URL.name == UriTypeEnum.URL.value


@pytest.mark.extra_test
def test_filename_from_uri():
    from doorway.x._uri import UriMalformedException
    from doorway.x._uri import uri_validate

    # test paths
    uri_validate("basename")
    uri_validate("basename.ext")
    uri_validate("basename.ext/path")
    uri_validate("/basename.ext/path")
    uri_validate("./basename.ext/path")
    uri_validate("/")
    uri_validate("./")
    uri_validate(".")
    with pytest.raises(
        UriMalformedException, match="field 'path' is required, but got value: None"
    ):
        uri_validate("")

    # test basic urls
    uri_validate("http://prefix/basename.ext/suffix")
    uri_validate("http://basename.ext/suffix")
    uri_validate("HTTP://basename.ext/suffix")
    with pytest.raises(
        UriMalformedException, match="field 'host' is required, but got value: None"
    ):
        uri_validate("http:/basename.ext/suffix")
    with pytest.raises(
        UriMalformedException, match="field 'host' is required, but got value: None"
    ):
        uri_validate("http:///basename.ext/suffix")

    # test url ports
    uri_validate("http://localhost:")
    uri_validate("http://localhost:/suffix")
    uri_validate("http://localhost:3000")
    uri_validate("http://localhost:3000/suffix")
    uri_validate("http://192.168.0.1")
    uri_validate("http://192.168.0.1:")
    uri_validate("http://192.168.0.1:3000")
    with pytest.raises(
        UriMalformedException, match="field 'host' is required, but got value: ''"
    ):
        uri_validate("http://:3000")

    # test urls and fragments etc
    uri_validate("http://basename.ext/suffix?query")
    uri_validate("http://basename.ext/suffix#fragment")
    uri_validate("http://basename.ext/suffix?query#fragment")
    uri_validate("http://basename.ext/suffix?query=5&query2=3#fragment")
    uri_validate("http://basename.ext/suffix;params?query=5&query2=3#fragment")
    uri_validate("http://basename.ext/suffix#fragment?query")
    uri_validate("http://basename.ext/suffix#fragment?query")


@pytest.mark.extra_test
def test_uri_paths_alt():
    from doorway.x._uri import uri_parse

    def uri(inp, targ=None):
        targ = inp if (targ is None) else targ
        assert uri_parse(inp).geturl() == targ

    uri(inp="path/uri_kind.ext")
    uri(inp="/path/uri_kind.ext")

    uri(inp="../path/uri_kind.ext")
    uri(inp="..//path/uri_kind.ext", targ="../path/uri_kind.ext")
    uri(inp="..//path//uri_kind.ext", targ="../path/uri_kind.ext")

    uri(inp="../../path//uri_kind.ext", targ="../../path/uri_kind.ext")
    uri(inp="../path/..//uri_kind.ext", targ="../uri_kind.ext")

    uri(inp="file:/path/uri_kind.ext")
    uri(inp="file:/path/uri_kind.ext")
    uri(inp="file:/./path/uri_kind.ext", targ="file:/path/uri_kind.ext")

    uri(inp="./path/uri_kind.ext", targ="path/uri_kind.ext")
    uri(inp="file:path/uri_kind.ext")
    uri(inp="file:./path/uri_kind.ext", targ="file:path/uri_kind.ext")
    uri(inp="file:.//path/uri_kind.ext", targ="file:path/uri_kind.ext")
    uri(inp="file:.//.//path/uri_kind.ext", targ="file:path/uri_kind.ext")

    uri(inp="file://path/uri_kind.ext", targ="file://path/uri_kind.ext")  # ERROR?
    uri(inp="file://path/uri_kind.ext")  # ERROR?
    uri(inp="file:///path/uri_kind.ext", targ="file:/path/uri_kind.ext")  # ERROR?
    uri(inp="file://./path/uri_kind.ext")  # ERROR?
    uri(inp="file:////path/uri_kind.ext", targ="file://path/uri_kind.ext")  # ERROR?
    uri(inp="file:///./path/uri_kind.ext", targ="file:/path/uri_kind.ext")  # ERROR?

    uri(inp="http://google.com/asdf")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

__all__ = (
    "EnvVar",
    # errors
    "EnvVarError",
    "EnvVarValidationError",
    "EnvVarConversionError",
    "EnvVarMissingError",
)

import os
from typing import Callable, Iterable, Optional
from typing import Sequence
from typing import Generic
from typing import TypeVar


# ========================================================================= #
# Types                                                                     #
# ========================================================================= #


T = TypeVar("T")
EnvVarFnConverterHint = Callable[[str], T]
EnvVarFnValidatorHint = Callable[[T], T]


# ========================================================================= #
# Errors                                                                    #
# ========================================================================= #


class EnvVarError(ValueError):
    """Base exception for environment variable errors."""

    pass


class EnvVarMissingError(EnvVarError):
    """Raised when a value is missing."""

    pass


class EnvVarValidationError(EnvVarError):
    """Raised when a value fails validation."""

    pass


class EnvVarConversionError(EnvVarError):
    """Raised when converting an environment variable string fails."""

    pass


# ========================================================================= #
# Variable Manager                                                          #
# ========================================================================= #


class EnvVar(Generic[T]):
    """
    A flexible handler for environment variables of any type.

    priority:
      0. validated force (if set, passed in)
      1. validated override (if set, from constructor)
      2. validated + converted environment variable (if set, from os.environ)
      3. validated default value (if set, passed in)
      4. validated fallback value (if set, from constructor)
      5. raise error if the value is missing (None)
    """

    def __init__(
        self,
        key: str,
        *,
        default: Optional[T] = None,
        converter: Optional[EnvVarFnConverterHint[T]] = None,  # applied to env str
        validator: Optional[EnvVarFnValidatorHint[T]] = None,  # applied to all values
    ):
        self._key = key
        self._env_converter = converter or (lambda x: x)
        self._val_validator = validator or (lambda x: x)
        # do not apply validator until get, more inefficient, but easier to test
        self._persisted_default = default  # value not passed through convert, only val
        self._persisted_override = None  # value not passed through convert, only val

    @property
    def env_key(self) -> str:
        return self._key

    # set or / clear with `None`

    def set_default_value(self, value: Optional[T]) -> None:
        # do not apply validator until get, more inefficient, but easier to test
        self._persisted_default = value

    def set_override_value(self, value: Optional[T]) -> None:
        # do not apply validator until get, more inefficient, but easier to test
        self._persisted_override = value

    # get

    @property
    def value(self) -> T:
        return self.get()

    def __call__(
        self, *, default: Optional[T] = None, override: Optional[T] = None
    ) -> T:
        return self.get(default=default, override=override)

    def get(self, *, default: Optional[T] = None, override: Optional[T] = None) -> T:
        # convert / extract values

        if override is not None:
            source = "call_override"
            result = override
        elif self._persisted_override is not None:
            source = "persisted_override"
            result = self._persisted_override  # already validated in constructor
        elif self._key in os.environ:
            source = "env_value"
            try:
                result = self._env_converter(os.environ[self._key])
            except Exception as e:
                raise EnvVarConversionError(
                    f"error getting {repr(self._key)} from os.environ, conversion failed: {e}"
                ) from e
        elif default is not None:
            source = "call_default"
            result = default
        elif self._persisted_default is not None:
            source = "persisted_default"
            result = self._persisted_default  # already validated in constructor
        else:
            raise EnvVarMissingError(f"error getting {repr(self._key)} from any source")
        # handle missing values
        if result is None:
            raise EnvVarMissingError(
                f"error getting {repr(self._key)} from {repr(source)}"
            )
        # validate the result
        try:
            return self._val_validator(result)
        except Exception as e:
            raise EnvVarValidationError(
                f"error getting {repr(self._key)} from {repr(source)}, validation failed: {e}"
            ) from e

    # ===== static helpers ===== #

    @classmethod
    def as_converter(cls, fn: EnvVarFnConverterHint[T]) -> EnvVarFnConverterHint[T]:
        def _converter(value: str) -> T:
            try:
                return fn(value)
            except Exception as e:
                raise EnvVarConversionError(f"error converting {value} to {fn}: {e}")

        return _converter

    @classmethod
    def as_validator(cls, fn: EnvVarFnValidatorHint[T]) -> EnvVarFnValidatorHint[T]:
        def _validator(value: T) -> T:
            try:
                return fn(value)
            except Exception as e:
                raise EnvVarValidationError(
                    f"error validating {repr(value)} with {fn}: {e}"
                )

        return _validator

    @classmethod
    def validator_sequence(
        cls,
        *fns: EnvVarFnValidatorHint[T],
    ) -> EnvVarFnValidatorHint[T]:
        def _validator(value: T) -> T:
            for fn in fns:
                if fn is not None:
                    value = fn(value)
            return value

        return _validator

    @classmethod
    def validator_allowed(cls, allowed_values: Iterable[T]) -> EnvVarFnValidatorHint[T]:
        allowed_values = set(allowed_values)
        if not allowed_values:
            raise ValueError("allowed_values must not be empty")

        def _validator(value: T) -> T:
            if value not in allowed_values:
                raise EnvVarValidationError(
                    f"value {repr(value)} must be one of: {allowed_values}"
                )
            return value

        return _validator

    @classmethod
    def validator_min_max(
        cls, min_value: Optional[T], max_value: Optional[T]
    ) -> EnvVarFnValidatorHint[T]:
        def _validator(value: T) -> T:
            if min_value is not None and value < min_value:
                raise EnvVarValidationError(
                    f"value {repr(value)} must be greater than or equal to {min_value}"
                )
            if max_value is not None and value > max_value:
                raise EnvVarValidationError(
                    f"value {repr(value)} must be less than or equal to {max_value}"
                )
            return value

        return _validator

    # ===== factory methods ===== #

    @classmethod
    def env_str(
        cls,
        key: str,
        *,
        default: Optional[str] = None,
        validator: Optional[EnvVarFnValidatorHint[str]] = None,
    ) -> "EnvVar[str]":
        return cls(
            key=key,
            default=default,
            converter=cls.as_converter(str),
            validator=validator,
        )

    @classmethod
    def env_int(
        cls,
        key: str,
        *,
        default: Optional[int] = None,
        validator: Optional[EnvVarFnValidatorHint[int]] = None,
    ) -> "EnvVar[int]":
        return cls(
            key=key,
            default=default,
            converter=cls.as_converter(int),
            validator=validator,
        )

    @classmethod
    def env_float(
        cls,
        key: str,
        *,
        default: Optional[float] = None,
        validator: Optional[EnvVarFnValidatorHint[float]] = None,
    ) -> "EnvVar[float]":
        return cls(
            key=key,
            default=default,
            converter=cls.as_converter(float),
            validator=validator,
        )

    @classmethod
    def env_bool(
        cls,
        key: str,
        *,
        default: Optional[bool] = None,
        # extra settings
        convert_keys_true: Sequence[str] = ("y", "yes", "t", "true", "1"),
        convert_keys_false: Sequence[str] = ("n", "no", "f", "false", "0"),
        convert_lowercase: bool = True,
        # not really useful for bool, but keep for consistency
        validator: Optional[EnvVarFnValidatorHint[bool]] = None,
    ) -> "EnvVar[bool]":
        def convert_bool(value: str) -> bool:
            if convert_lowercase:
                value = value.lower()
            if value in convert_keys_true:
                return True
            elif value in convert_keys_false:
                return False
            else:
                raise EnvVarConversionError(
                    f"cannot convert environment variable `{key}={value}` into bool, must be one of: {sorted(list(convert_keys_true) + list(convert_keys_false))}"
                )

        return cls(
            key=key,
            default=default,
            converter=convert_bool,
            validator=validator,
        )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


__all__ = [
    "shard_hash",
    "shard_idx",
    "sharded",
    "sharded_weighted",
]

import os
from pathlib import Path
from typing import Callable, List, Literal, Tuple
from typing import Iterable
from typing import Optional
from typing import Union
from typing import TypeVar

from doorway._hash import Hash
from doorway._hash import HashAlgo
from doorway._hash import hash_str


# ========================================================================= #
# individual shards                                                         #
# ========================================================================= #


T = TypeVar("T")

_SHARD_KEYS = {
    "basename": os.path.basename,
    "abspath": os.path.abspath,
    "input": lambda x: x,
}

ShardKey = Optional[Union[str, Callable[[T], str]]]


def shard_hash(
    value: T,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> Hash:
    # get the hash data function
    if shard_key is None:
        pass
    elif callable(shard_key):
        value = shard_key(value)
    elif isinstance(shard_key, str):
        fn = _SHARD_KEYS.get(shard_key, None)  # cannot be None here
        if fn is None:
            raise KeyError(
                f"if shard_key is a str, it must be one of: {list(_SHARD_KEYS.keys())}, got: {repr(shard_key)}"
            )
        value = fn(value)
    else:
        raise ValueError(
            f"shard_key must be a str, callable or None, got: {repr(shard_key)}"
        )
    # get the string
    assert isinstance(value, (str, Path)), (
        f"The value after shard_key is applied must be a str or Path, instead got type: {type(value)}, with value: {repr(value)}"
    )
    # compute the hash
    return hash_str(str(value), hash_algo=hash_algo)


def shard_idx(
    value: T,
    num_shards: int,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> int:
    assert isinstance(num_shards, int) and (num_shards > 0), (
        f"num_shards must be an integer that is > 0, got: {repr(num_shards)}"
    )
    # compute the hash for the value
    hash = shard_hash(value, shard_key=shard_key, hash_algo=hash_algo)
    # convert hashes to integers, and assign to correct split
    return int(hash, 16) % num_shards


# ========================================================================= #
# multiples shards                                                          #
# ========================================================================= #


_SHARD_RETURNS = {
    "pairs": lambda i, value: (i, value),
    "indices": lambda i, value: i,
    "values": lambda i, value: value,
}

_ShardsReturnHint = Union[
    List[Tuple[int, T]],  # pairs
    List[int],  # indices
    List[T],  # values
]


def sharded(
    values: Iterable[T],
    num_shards: int,
    *,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
    returns: Literal["pairs", "indices", "values"] = "values",
) -> _ShardsReturnHint[T]:
    """
    Shard files based on their hashes instead of random seeds
    """
    # shard functions
    if returns not in _SHARD_RETURNS:
        raise KeyError(
            f"invalid shards returns: {repr(returns)}, must be one of: {sorted(_SHARD_RETURNS.keys())}"
        )
    value_getter = _SHARD_RETURNS[returns]
    # create new array of shards
    shards = [[] for _ in range(num_shards)]
    # assign paths to shards
    for i, value in enumerate(values):
        idx = shard_idx(value, num_shards, shard_key=shard_key, hash_algo=hash_algo)
        shards[idx].append(value_getter(i, value))
    # results
    return shards


def sharded_weighted(
    values: Iterable[T],
    shard_weights: Iterable[int],
    *,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
    returns: Literal["pairs", "indices", "values"] = "values",
) -> _ShardsReturnHint[T]:
    """
    Shard files based on their hashes instead of random seeds, and group them based on weights
    -- This is useful if you need to split a dataset, but you expect changes to be made to it,
       eg. files will always randomly be assigned to the same shared
    """
    shard_weights = list(shard_weights)
    # checks
    assert all(isinstance(size, int) and (size >= 0) for size in shard_weights), (
        f"values of group_sizes must be integers that are >= 0, got: {repr(shard_weights)}"
    )
    # get all the shards
    shards = sharded(
        values=values,
        num_shards=sum(shard_weights),
        shard_key=shard_key,
        hash_algo=hash_algo,
        returns=returns,
    )
    # group all the shards together
    weighted_buckets, i = [], 0
    for num_shards in shard_weights:
        weighted_buckets.append(
            [item for shard in shards[i : i + num_shards] for item in shard]
        )
        i += num_shards
    # done!
    return weighted_buckets


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

__all__ = (
    "RST",
    # dark colors
    "GRY",
    "lRED",
    "lGRN",
    "lYLW",
    "lBLU",
    "lMGT",
    "lCYN",
    "WHT",
    # light colors
    "BLK",
    "RED",
    "GRN",
    "YLW",
    "BLU",
    "MGT",
    "CYN",
    "lGRY",
)


# ========================================================================= #
# Ansi Colors                                                               #
# ========================================================================= #


RST = "\033[0m"

# dark colors
GRY = "\033[90m"
lRED = "\033[91m"
lGRN = "\033[92m"
lYLW = "\033[93m"
lBLU = "\033[94m"
lMGT = "\033[95m"
lCYN = "\033[96m"
WHT = "\033[97m"

# light colors
BLK = "\033[30m"
RED = "\033[31m"
GRN = "\033[32m"
YLW = "\033[33m"
BLU = "\033[34m"
MGT = "\033[35m"
CYN = "\033[36m"
lGRY = "\033[37m"


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

__all__ = [
    "io_download",
]

import logging
import os
import warnings
from typing import Literal


from doorway._atomic import AtomicOpen


LOG = logging.getLogger(__name__)


# ========================================================================= #
# files/dirs exist                                                          #
# ========================================================================= #


def io_download(
    src_url: str,
    dst_path: str,
    *,
    exists_mode: Literal["skip", "overwrite", "error"] = "error",
    chunk_size: int = 16384,
    progress: bool = True,
):
    # make sure we have the correct imports
    try:
        import requests
    except ImportError:
        raise ImportError(
            "The `requests` package is required for downloading files.\n"
            "You can install it via: `pip install requests`."
        )

    try:
        from tqdm import tqdm
    except ImportError:
        warnings.warn(
            "The `tqdm` package is not installed, progress bar will not be shown.\n"
        )
        tqdm = None
        progress = False

    # skip existing
    if exists_mode == "skip":
        if os.path.exists(dst_path):
            LOG.info(f"Skipping: {dst_path}")
            return
        overwrite = False
    elif exists_mode == "overwrite":
        overwrite = True
    elif exists_mode == "error":
        overwrite = False
    else:
        raise ValueError(f"invalid `exists_mode`: {exists_mode}")

    # write the file
    with AtomicOpen(dst_path, "wb" if overwrite else "xb") as fp:
        response = requests.get(src_url, stream=True)

        # get the file size from the request for the progress bar
        total_length = response.headers.get("content-length")
        if total_length is not None:
            total_length = int(total_length)

        # download with progress bar
        LOG.info(f"Downloading: {src_url} to: {dst_path}")

        if progress and tqdm is not None:
            with tqdm(
                total=total_length,
                desc="Downloading",
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as progress:
                for data in response.iter_content(chunk_size=chunk_size):
                    fp.write(data)
                    progress.update(chunk_size)
        else:
            for data in response.iter_content(chunk_size=chunk_size):
                fp.write(data)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

__all__ = [
    "c",
    # ===== atomic ===== #
    "AtomicPath",
    "AtomicOpen",
    # ===== context ===== #
    "ctx_no_stdout",
    "ctx_no_stderr",
    "ctx_temp_attr",
    "ctx_temp_wd",
    "ctx_temp_sys_args",
    "ctx_temp_environ",
    "ctx_do_undo",
    # ===== env_vars ===== #
    "EnvVar",
    # errors
    "EnvVarError",
    "EnvVarMissingError",
    "EnvVarValidationError",
    "EnvVarConversionError",
    # ===== fmt ===== #
    "fmt_use_colors_get",
    "fmt_bytes_to_human",
    # ===== hash ===== #
    # types
    "Hash",
    "Hashes",
    "HashMode",
    "HashAlgo",
    "HashPath",
    # errors
    "HashError",
    # hash mode
    "hash_mode_get",
    "hash_algo_get",
    # normalise hash
    "hash_norm",
    # compute hash
    "hash_bytes",
    "hash_bytes_iter",
    "hash_str",
    "hash_file",
    "hash_file_validate",
    "hash_file_is_valid",
    # ===== inout ===== #
    "io_download",
    # ===== modify_path ===== #
    "basename_split_ext",
    "basename_modify",
    "path_basename_modify",
    # ===== shard ===== #
    "shard_hash",
    "shard_idx",
    "sharded",
    "sharded_weighted",
    # ===== stale ===== #
    "stalefile_is_stale",
    "stalefile_generate",
    "stalefile_decorator",
    "Stalefile",
]

# colors
import doorway._colors as c

# other
from doorway._atomic import (
    AtomicPath,
    AtomicOpen,
)
from doorway._ctx import (
    ctx_no_stdout,
    ctx_no_stderr,
    ctx_temp_attr,
    ctx_temp_wd,
    ctx_temp_sys_args,
    ctx_temp_environ,
    ctx_do_undo,
)
from doorway._env_vars import (
    EnvVar,
    EnvVarError,
    EnvVarValidationError,
    EnvVarMissingError,
    EnvVarConversionError,
)
from doorway._fmt import (
    fmt_use_colors_get,
    fmt_bytes_to_human,
)
from doorway._hash import (
    Hash,
    Hashes,
    HashMode,
    HashAlgo,
    HashPath,
    HashError,
    hash_mode_get,
    hash_algo_get,
    hash_norm,
    hash_bytes,
    hash_bytes_iter,
    hash_str,
    hash_file,
    hash_file_validate,
    hash_file_is_valid,
)
from doorway._inout import (
    io_download,
)
from doorway._modify_path import (
    basename_split_ext,
    basename_modify,
    path_basename_modify,
)
from doorway._shard import (
    shard_hash,
    shard_idx,
    sharded,
    sharded_weighted,
)
from doorway._stale import (
    stalefile_is_stale,
    stalefile_generate,
    stalefile_decorator,
    Stalefile,
)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2025 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


__all__ = [
    "fmt_use_colors_get",
    "fmt_bytes_to_human",
]


import math
from typing import Optional

import doorway._colors as c
from doorway._env_vars import EnvVar


# ========================================================================= #
# Variable Handlers                                                         #
# ========================================================================= #


_VAR_HANDLER_USE_COLORS = EnvVar.env_bool(
    key="DOORWAY_ENABLE_COLORS",
    default=True,
)


def fmt_use_colors_get(use_colors: Optional[bool] = None) -> bool:
    return _VAR_HANDLER_USE_COLORS.get(override=use_colors)


# ========================================================================= #
# Byte Formatting                                                           #
# ========================================================================= #


# the names and colours of the different units
_BYTES_UNIT_COLORS = (
    c.WHT,
    c.lGRN,
    c.lYLW,
    c.lRED,
    c.lRED,
    c.lRED,
    c.lRED,
    c.lRED,
    c.lRED,
)
_BYTES_UNIT_NAMES = {
    1024: ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB", "RiB", "QiB"),
    1000: ("B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "RB", "QB"),
}

# amount of extra padding to add to the left of the dot
_BYTES_BASE_PADDING = {
    1024: 5,
    1000: 4,
}

_BYTES_UNIT_PADDING = {
    1024: 3,
    1000: 2,
}


def fmt_bytes_to_human(
    size_bytes: int,
    *,
    base: int = 1024,
    decimals: int = 3,
    align: bool = False,
    use_colors: Optional[bool] = None,
    round_unit: bool = True,
) -> str:
    """
    Obtain the human-readable string representation of the given bytes

    SI Units - base 1000
        k   kilo   1000**1  !!! lowercase k is used for kilo, because uppercase K is used for kelvin
        M   mega   1000**2
        G   giga   1000**3
        T   tera   1000**4
        P   peta   1000**5
        E   exa    1000**6
        Z   zetta  1000**7
        Y   yotta  1000**8
        R   ronna  1000**9
        Q   quetta 1000**10

    IEC Units - base 1024
        Symbol	Prefix	Example	Size
        Ki kibi  kibibyte  (KiB) 2**10  (1024**1)
        Mi mebi  mebibyte  (MiB) 2**20  (1024**2)
        Gi gibi  gibibyte  (GiB) 2**30  (1024**3)
        Ti tebi  tebibyte  (TiB) 2**40  (1024**4)
        Pi pebi  pebibyte  (PiB) 2**50  (1024**5)
        Ei exbi  exbibyte  (EiB) 2**60  (1024**6)
        Zi zebi  zebibyte  (ZiB) 2**70  (1024**7)
        Yi yobi  yobibyte  (YiB) 2**80  (1024**8)
        Ri robi  robibyte  (RiB) 2**90  (1024**9)
        Qi quebi quebibyte (QiB) 2**100 (1024**10)
    """
    # check the unit of measurement
    if not isinstance(size_bytes, int):
        raise TypeError(
            f"invalid size in bytes, must be of type `int`, got: {type(size_bytes)}"
        )
    if not isinstance(base, int):
        raise TypeError(
            f"invalid bytes base number, must be of type `int`, got: {type(base)}"
        )
    if size_bytes < 0:
        raise ValueError(f"invalid size in bytes, cannot be negative: {size_bytes}")
    if base not in _BYTES_UNIT_NAMES:
        raise ValueError(
            f"invalid bytes base number: {repr(base)} must be one of: {sorted(_BYTES_UNIT_NAMES.keys())}"
        )
    units = _BYTES_UNIT_NAMES[base]

    # 1. compute power
    # NOTE: This is not precise, we should rather be using Fractions and then
    #       instead of using `math.log`, rather compute the multiples of the
    #       bases and then checking if the size_bytes is in the required range
    power = 0 if (size_bytes == 0) else int(math.log(size_bytes, base))
    # bound the power by the maximum available unit
    i = min(power, len(units) - 1)

    # 2. compute formatted unit by dividing
    # NOTE: divide in integer space to avoid precision errors, this is
    #       a floor operation, so we never round up to the next unit,
    #       ie.  `fmt_bytes_to_human(1024**8 - 1, base=1024) == "1023.999 ZiB"`
    #       with this, we don't need to round and update the unit below:
    #       `size_fmt = size_bytes // (base**max(0, i-1)) / (base**min(1, i))`
    size_fmt = size_bytes / (base**i)

    # 3. round the formatted unit and update if the unit changes
    if round_unit:
        size_fmt = round(size_fmt, decimals)
        if (size_fmt >= base) and (i < len(units) - 1):
            size_fmt = round(size_fmt / base, decimals)
            i += 1

    # obtain the actual unit strings
    unit = units[i]
    if fmt_use_colors_get(use_colors):
        unit = f"{_BYTES_UNIT_COLORS[i]}{unit}{c.RST}"

    # format string
    if align:
        lpad = _BYTES_BASE_PADDING[base]
        rpad = _BYTES_UNIT_PADDING[base]
        return f"{size_fmt:>{lpad + decimals}.{decimals}f} {unit:<{rpad}s}"
    else:
        return f"{size_fmt:.{decimals}f} {unit}"


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import itertools
import tempfile
from contextlib import contextmanager
from contextlib import nullcontext
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional

import pytest

import doorway._hash
from doorway._hash import hash_mode_get
from doorway._hash import hash_algo_get
from doorway._env_vars import EnvVar, EnvVarMissingError, EnvVarValidationError
from doorway._ctx import ctx_temp_attr
from doorway._ctx import ctx_temp_environ


# ========================================================================= #
# TEST UTILS                                                                #
# ========================================================================= #


@contextmanager
def skip_context(value, ctx):
    if value is not None:
        with ctx:
            yield
    else:
        yield


def context_temp_hash_mode_default(
    hash_mode: Optional[str], target=doorway._hash._VAR_HANDLER_HASH_MODE
):
    return skip_context(
        hash_mode, ctx_temp_attr(target, "_persisted_default", hash_mode)
    )


# ========================================================================= #
# TEST VAR HANDLER                                                          #
# ========================================================================= #


INVALID = "<INVALID>"


@pytest.mark.parametrize(
    ("usr_override", "cls_override", "environ", "usr_default", "cls_default"),
    itertools.product(
        ["a", INVALID, None],  # usr_override
        ["b", INVALID, None],  # cls_override
        ["c", INVALID, None],  # environ
        ["d", INVALID, None],  # usr_default
        ["e", INVALID, None],  # cls_default
    ),
)
def test_get_hash_mode(
    usr_override: str,
    cls_override: str,
    environ: str,
    usr_default: str,
    cls_default: str,
):
    # forcefully set the fallback, environ, and defaults
    with skip_context(environ, ctx_temp_environ(ASDF_FDSA=environ)):
        env_var = EnvVar.env_str(
            key="ASDF_FDSA",
            validator=EnvVar.validator_allowed(["a", "b", "c", "d", "e"]),
        )
        env_var.set_default_value(cls_default)
        env_var.set_override_value(cls_override)
        # now get
        values = [usr_override, cls_override, environ, usr_default, cls_default]
        first = next((v for v in values if v is not None), None)

        if first is None:
            with pytest.raises(EnvVarMissingError):
                env_var.get(default=usr_default, override=usr_override)
        elif first == INVALID:
            with pytest.raises(EnvVarValidationError):
                env_var.get(default=usr_default, override=usr_override)
        else:
            assert env_var.get(default=usr_default, override=usr_override) == first


def test_hash_mode_set_default():
    assert hash_mode_get() == "fast"
    # check that environ doesnt overwrite & that setting to null resets
    with ctx_temp_environ(DOORWAY_HASH_MODE="full"):
        assert hash_mode_get() == "full"
    assert hash_mode_get() == "fast"
    # check invalid
    with ctx_temp_environ(DOORWAY_HASH_MODE="INVALID"):
        with pytest.raises(EnvVarValidationError):
            hash_mode_get()


def test_hash_algo_set_default():
    assert hash_algo_get() == "md5"
    # check that environ doesnt overwrite & that setting to null resets
    with ctx_temp_environ(DOORWAY_HASH_ALGO="sha1"):
        assert hash_algo_get() == "sha1"
    assert hash_algo_get() == "md5"
    # check invalid
    with ctx_temp_environ(DOORWAY_HASH_ALGO="INVALID"):
        with pytest.raises(EnvVarValidationError):
            hash_algo_get()


def test_hash_norm():
    assert doorway.hash_norm("???") == "???"
    assert (
        doorway.hash_norm({"fast:md5": "fast:md5", "fast": "fast", "md5": "md5"})
        == "fast:md5"
    )
    assert doorway.hash_norm({"fast:md5": "fast:md5", "fast": "fast"}) == "fast:md5"
    assert doorway.hash_norm({"fast:md5": "fast:md5", "md5": "md5"}) == "fast:md5"
    assert doorway.hash_norm({"fast:md5": "fast:md5"}) == "fast:md5"
    assert doorway.hash_norm({"fast": "fast", "md5": "md5"}) == "fast"
    assert doorway.hash_norm({"fast": "fast"}) == "fast"
    assert doorway.hash_norm({"md5": "md5"}) == "md5"
    with pytest.raises(
        KeyError,
        match="hash dictionary does not contain a valid key for either 1. 'fast:md5', 2. 'fast', or 3. 'md5'. Available hash keys are: \\[\\]",
    ):
        assert doorway.hash_norm({})
    # check overrides 1.
    assert (
        doorway.hash_norm({"fast:md5": "fast:md5"}, hash_mode=None, hash_algo=None)
        == "fast:md5"
    )
    assert doorway.hash_norm({"fast": "fast"}, hash_mode=None, hash_algo=None) == "fast"
    assert doorway.hash_norm({"md5": "md5"}, hash_mode=None, hash_algo=None) == "md5"
    with pytest.raises(
        KeyError,
        match="hash dictionary does not contain a valid key for either 1. 'fast:md5', 2. 'fast', or 3. 'md5'. Available hash keys are: \\['invalid'\\]",
    ):
        doorway.hash_norm({"invalid": "invalid"}, hash_mode=None, hash_algo=None)
    # check overrides 2.
    assert (
        doorway.hash_norm({"full:md5": "full:md5"}, hash_mode="full", hash_algo=None)
        == "full:md5"
    )
    assert (
        doorway.hash_norm({"full": "full"}, hash_mode="full", hash_algo=None) == "full"
    )
    assert doorway.hash_norm({"md5": "md5"}, hash_mode="full", hash_algo=None) == "md5"
    with pytest.raises(
        KeyError,
        match="hash dictionary does not contain a valid key for either 1. 'full:md5', 2. 'full', or 3. 'md5'. Available hash keys are: \\['invalid'\\]",
    ):
        doorway.hash_norm({"invalid": "invalid"}, hash_mode="full", hash_algo=None)
    # check overrides 3.
    assert (
        doorway.hash_norm({"fast:sha1": "fast:sha1"}, hash_mode=None, hash_algo="sha1")
        == "fast:sha1"
    )
    assert (
        doorway.hash_norm({"fast": "fast"}, hash_mode=None, hash_algo="sha1") == "fast"
    )
    assert (
        doorway.hash_norm({"sha1": "sha1"}, hash_mode=None, hash_algo="sha1") == "sha1"
    )
    with pytest.raises(
        KeyError,
        match="hash dictionary does not contain a valid key for either 1. 'fast:sha1', 2. 'fast', or 3. 'sha1'. Available hash keys are: \\['invalid'\\]",
    ):
        doorway.hash_norm({"invalid": "invalid"}, hash_mode=None, hash_algo="sha1")
    # check overrides 4.
    assert (
        doorway.hash_norm(
            {"full:sha1": "full:sha1"}, hash_mode="full", hash_algo="sha1"
        )
        == "full:sha1"
    )
    assert (
        doorway.hash_norm({"full": "full"}, hash_mode="full", hash_algo="sha1")
        == "full"
    )
    assert (
        doorway.hash_norm({"sha1": "sha1"}, hash_mode="full", hash_algo="sha1")
        == "sha1"
    )
    with pytest.raises(
        KeyError,
        match="hash dictionary does not contain a valid key for either 1. 'full:sha1', 2. 'full', or 3. 'sha1'. Available hash keys are: \\['invalid'\\]",
    ):
        doorway.hash_norm({"invalid": "invalid"}, hash_mode="full", hash_algo="sha1")


def test_hash_norm_invalid():
    with pytest.raises(
        TypeError,
        match="normalized hash should be a str, got type: <class 'NoneType'> for value: None",
    ):
        assert doorway.hash_norm(None)
    with pytest.raises(
        TypeError,
        match="normalized hash should be a str, got type: <class 'int'> for value: 1",
    ):
        assert doorway.hash_norm(1)
    with pytest.raises(
        TypeError,
        match="normalized hash should be a str, got type: <class 'NoneType'> for value: None",
    ):
        assert doorway.hash_norm({"fast:md5": None})
    with pytest.raises(
        TypeError,
        match="normalized hash should be a str, got type: <class 'int'> for value: 1",
    ):
        assert doorway.hash_norm({"fast:md5": 1})


# ========================================================================= #
# TEST HASHING - HELPER                                                     #
# ========================================================================= #


def _make_sequential_file(file, length: int = 100_000):
    context = open(file, "w") if isinstance(file, (str, Path)) else nullcontext(file)
    # insert lines into the file
    with context as fp:
        for i in range(length):
            fp.write(f"{i}\n")
    fp.seek(0)
    return fp


# ========================================================================= #
# TEST HASHING                                                              #
# ========================================================================= #


def test_hash_file_dir():
    with tempfile.TemporaryDirectory() as d:
        with pytest.raises(
            IsADirectoryError, match="the path exists but is not a file:"
        ):
            doorway.hash_file(d, hash_algo="md5", hash_mode="fast", hash_missing=False)
    with pytest.raises(
        FileNotFoundError, match="could not compute hash for missing file:"
    ):
        doorway.hash_file(d, hash_algo="md5", hash_mode="fast", hash_missing=False)


def test_hash_file():
    with NamedTemporaryFile("w") as fp:
        fp = _make_sequential_file(fp, 1_000_000)
        # hash the file, checking default vars
        assert doorway.hash_file(fp.name) == "f71b9a144c7a67c43999103f34c5a0ef"
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode=None, hash_missing=False
            )
            == "f71b9a144c7a67c43999103f34c5a0ef"
        )
        with context_temp_hash_mode_default("full"):
            assert doorway.hash_file(fp.name) == "762251ff53a76f10ada68131f8e3d4c1"
        # hash the file checking explicit modes
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="fast", hash_missing=False
            )
            == "f71b9a144c7a67c43999103f34c5a0ef"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="full", hash_missing=False
            )
            == "762251ff53a76f10ada68131f8e3d4c1"
        )
        # hash the file checking explicit types -- might be an idea to swap to sha256 by default instead?
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha1", hash_mode="fast", hash_missing=False
            )
            == "c1ae8652e374a052493c1d7faae28f41c139c87c"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha1", hash_mode="full", hash_missing=False
            )
            == "5e174204d2aae9c9adf7e11b2184f36a56730230"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha256", hash_mode="fast", hash_missing=False
            )
            == "b0cab88abcf70e9cf2fac9db6eccf0141c818ded4888880936169b7f88b37fa6"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha256", hash_mode="full", hash_missing=False
            )
            == "7b8f269ab1f1ba01ea1cb69d69eb2abdd98b88311ce896f1083cc9e66112988b"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha512", hash_mode="fast", hash_missing=False
            )
            == "c3e92e686f010bed0b9fcf404e87fea6fb049bae6bb2b2a12c0d45c0b686caa39c91f2e6d792731d70393f07757e2ce1753be175dae287fa991b5c23e2d7ae69"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="sha512", hash_mode="full", hash_missing=False
            )
            == "b1ad0bfbe6a5560623fc66926ec63a3c11856f505bdbdd713a608d18bbb32b1aaa7260d91558cdf1cb8bd5bc3e539c51b25badd8bf3229f2a9906dbdcb29c7d5"
        )
    # test missing files
    with pytest.raises(
        FileNotFoundError, match="could not compute hash for missing file"
    ):
        doorway.hash_file(fp.name)
    with pytest.raises(
        FileNotFoundError, match="could not compute hash for missing file"
    ):
        doorway.hash_file(
            fp.name, hash_algo="md5", hash_mode="fast", hash_missing=False
        )
    assert (
        doorway.hash_file(fp.name, hash_algo="md5", hash_mode="fast", hash_missing=True)
        == ""
    )


def test_hash_file_small():
    # check tiny file
    with NamedTemporaryFile("w") as fp:
        fp = _make_sequential_file(fp, 100)
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="fast", hash_missing=False
            )
            == "40f6ddb8db1f93ad1f5502e2e0321f2f"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="full", hash_missing=False
            )
            == "9a10f4f09341c2db76a16b44f841c551"
        )
    # check mini file
    with NamedTemporaryFile("w") as fp:
        fp = _make_sequential_file(fp, 1000)
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="fast", hash_missing=False
            )
            == "c8902cdef4e5ad7c0b59784ebe454aa9"
        )
        assert (
            doorway.hash_file(
                fp.name, hash_algo="md5", hash_mode="full", hash_missing=False
            )
            == "b6f42041b389b22d1fb65ec3f1307ccd"
        )


def test_hash_file_validate():
    hashs_md5 = {
        "fast": "f71b9a144c7a67c43999103f34c5a0ef",
        "full": "762251ff53a76f10ada68131f8e3d4c1",
    }
    hashs_sha1 = {
        "fast": "c1ae8652e374a052493c1d7faae28f41c139c87c",
        "full": "5e174204d2aae9c9adf7e11b2184f36a56730230",
    }
    # get hashes
    with NamedTemporaryFile("w") as fp:
        fp = _make_sequential_file(fp, 1_000_000)
        # check validation of file
        doorway.hash_file_validate(fp.name, hash=hashs_md5["fast"])
        doorway.hash_file_validate(
            fp.name,
            hash=hashs_md5["fast"],
            hash_algo="md5",
            hash_mode="fast",
            hash_missing=False,
        )
        # check hash dictionaries
        doorway.hash_file_validate(
            fp.name,
            hash=hashs_md5,
            hash_algo="md5",
            hash_mode="fast",
            hash_missing=False,
        )
        doorway.hash_file_validate(
            fp.name,
            hash=hashs_md5,
            hash_algo="md5",
            hash_mode="full",
            hash_missing=False,
        )
        # sha1
        doorway.hash_file_validate(
            fp.name,
            hash=hashs_sha1,
            hash_algo="sha1",
            hash_mode="fast",
            hash_missing=False,
        )
        doorway.hash_file_validate(
            fp.name,
            hash=hashs_sha1,
            hash_algo="sha1",
            hash_mode="full",
            hash_missing=False,
        )
        # check changing hash mode
        with context_temp_hash_mode_default("full"):
            doorway.hash_file_validate(fp.name, hash=hashs_md5["full"])
            with pytest.raises(
                doorway.HashError,
                match="computed full md5 hash: '762251ff53a76f10ada68131f8e3d4c1' does not match expected hash: 'f71b9a144c7a67c43999103f34c5a0ef' for file:",
            ):
                doorway.hash_file_validate(fp.name, hash=hashs_md5["fast"])
        # check changing hash mode
        with context_temp_hash_mode_default("fast"):
            doorway.hash_file_validate(fp.name, hash=hashs_md5["fast"])
            with pytest.raises(
                doorway.HashError,
                match="computed fast md5 hash: 'f71b9a144c7a67c43999103f34c5a0ef' does not match expected hash: '762251ff53a76f10ada68131f8e3d4c1' for file:",
            ):
                doorway.hash_file_validate(fp.name, hash=hashs_md5["full"])
        # check invalid hash
        with pytest.raises(
            doorway.HashError,
            match="computed fast md5 hash: 'f71b9a144c7a67c43999103f34c5a0ef' does not match expected hash: '<invalid>' for file:",
        ):
            doorway.hash_file_validate(fp.name, hash="<invalid>", hash_algo="md5")
        # check missing hash
        with pytest.raises(
            KeyError,
            match="hash dictionary does not contain a valid key for either 1. 'fast:md5', 2. 'fast', or 3. 'md5'. Available hash keys are: \[\]",
        ):
            doorway.hash_file_validate(fp.name, hash={}, hash_algo="md5")
        # check is valid
        assert doorway.hash_file_is_valid(fp.name, hash=hashs_md5, hash_missing=False)
        assert doorway.hash_file_is_valid(
            fp.name, hash=hashs_md5["fast"], hash_missing=False
        )
        assert not doorway.hash_file_is_valid(
            fp.name, hash=hashs_md5["full"], hash_missing=False
        )
        assert doorway.hash_file_is_valid(fp.name, hash=hashs_md5, hash_missing=True)
        assert doorway.hash_file_is_valid(
            fp.name, hash=hashs_md5["fast"], hash_missing=True
        )
        assert not doorway.hash_file_is_valid(
            fp.name, hash=hashs_md5["full"], hash_missing=True
        )
    # missing file
    with pytest.raises(
        FileNotFoundError, match="could not compute hash for missing file:"
    ):
        doorway.hash_file_validate(fp.name, hash=hashs_md5["fast"], hash_missing=False)
    with pytest.raises(
        doorway.HashError,
        match="computed fast md5 hash: '' does not match expected hash: 'f71b9a144c7a67c43999103f34c5a0ef' for file:",
    ):
        doorway.hash_file_validate(fp.name, hash=hashs_md5["fast"], hash_missing=True)
    # missing file
    with pytest.raises(
        FileNotFoundError, match="could not compute hash for missing file:"
    ):
        doorway.hash_file_is_valid(fp.name, hash=hashs_md5, hash_missing=False)
    assert not doorway.hash_file_is_valid(fp.name, hash=hashs_md5, hash_missing=True)
    assert not doorway.hash_file_is_valid(
        fp.name, hash=hashs_md5["fast"], hash_missing=True
    )
    assert not doorway.hash_file_is_valid(
        fp.name, hash=hashs_md5["full"], hash_missing=True
    )


# ========================================================================= #
# Variable Handler                                                          #
# ========================================================================= #


def test_variable_handler():
    handler = EnvVar.env_str(
        key="VAR_HANDLER",
        default="1",
        validator=EnvVar.validator_allowed(["1", "2", "3"]),
    )
    assert handler.env_key == "VAR_HANDLER"
    # environ values
    assert handler.get() == "1"
    with ctx_temp_environ(VAR_HANDLER="3"):
        assert handler.get() == "3"
    assert handler.get() == "1"

    # checks
    handler.set_default_value("2")
    assert handler.get() == "2"
    handler.set_default_value("2")
    assert handler.get() == "2"

    handler.set_default_value(None)
    with pytest.raises(EnvVarMissingError):
        handler.get()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
